/* Dazuko FreeBSD. Allow FreeBSD file access control for 3rd-party applications.
   Written by John Ogness <jogness@antivir.de>

   Copyright (C) 2003, H+BEDV Datentechnik GmbH
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. Neither the name of Dazuko nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/


#include <sys/types.h>
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/proc.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/syscall.h>
#include <machine/stdarg.h>

#include "dazuko_freebsd.h"
#include "dazuko_xp.h"
#include "dazukoio.h"


d_read_t freebsd_dazuko_device_read;
d_ioctl_t freebsd_dazuko_device_ioctl;
d_open_t freebsd_dazuko_device_open;
d_close_t freebsd_dazuko_device_close;

extern struct xp_atomic active;

static int	dev_major = 33;
static dev_t	sdev = NULL;

#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_open(struct proc *p, struct open_args *uap);
static struct sysent freebsd_dazuko_sys_open_sysent = { 3, (sy_call_t*)freebsd_dazuko_sys_open };
static sy_call_t *original_sys_open;
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_close(struct proc *p, struct close_args *uap);
static struct sysent freebsd_dazuko_sys_close_sysent = { 1, (sy_call_t*)freebsd_dazuko_sys_close };
static sy_call_t *original_sys_close;
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
static int freebsd_dazuko_sys_write(struct proc *p, struct write_args *uap);
static struct sysent freebsd_dazuko_sys_write_sysent = { 3, (sy_call_t*)freebsd_dazuko_sys_write };
static sy_call_t *original_sys_write;
#endif

#ifdef ON_EXEC_SUPPORT
static int freebsd_dazuko_sys_execve(struct proc *p, struct execve_args *uap);
static struct sysent freebsd_dazuko_sys_execve_sysent = { 3, (sy_call_t*)freebsd_dazuko_sys_execve };
static sy_call_t *original_sys_execve;
#endif   

static struct cdevsw	cdepts = {
					freebsd_dazuko_device_open,
					freebsd_dazuko_device_close,
					freebsd_dazuko_device_read,
					nowrite,
					freebsd_dazuko_device_ioctl,
					nopoll,
					nommap,
					nostrategy,
					DEVICE_NAME,
					33,
					nodump,
					nopsize,
					D_TTY,
					-1
				};


MALLOC_DECLARE(M_DAZUKOBUF);
MALLOC_DEFINE(M_DAZUKOBUF, "dazukobuffer", "buffer for dazuko module");


static inline int freebsd_getpid(struct proc *p)
{
	if (p == NULL)
	{
		xp_print("dazuko: warning: freebsd_getpid(NULL)\n");
		return -1;
	}

	return p->p_pid;
}


/* mutex */

inline int xp_init_mutex(struct xp_mutex *mutex)
{
	/* not implemented */
	return 0;
}

inline int xp_down(struct xp_mutex *mutex)
{
	/* not implemented */
	return 0;
}

inline int xp_up(struct xp_mutex *mutex)
{
	/* not implemented */
	return 0;
}


/* read-write lock */

inline int xp_init_rwlock(struct xp_rwlock *rwlock)
{
	/* not implemented */
	return 0;
}

inline int xp_write_lock(struct xp_rwlock *rwlock)
{
	/* not implemented */
	return 0;
}

inline int xp_write_unlock(struct xp_rwlock *rwlock)
{
	/* not implemented */
	return 0;
}

inline int xp_read_lock(struct xp_rwlock *rlock)
{
	/* not implemented */
	return 0;
}

inline int xp_read_unlock(struct xp_rwlock *rlock)
{
	/* not implemented */
	return 0;
}


/* wait-notify queue */

inline int xp_init_queue(struct xp_queue *queue)
{
	/* not implemented */
	return 0;
}

inline int xp_wait_until_condition(struct xp_queue *queue, int (*cfunction)(void *), void *cparam)
{
	while (1)
	{
		if (asleep(queue, PCATCH, "dazuko", 0) != 0)
			return EINTR;

		if (cfunction(cparam))
			break;

		if (await(PCATCH, 0) != 0)
			return EINTR;
	}

	asleep(NULL, 0, "", 0);

	return 0;
}

inline int xp_notify(struct xp_queue *queue)
{
	wakeup(queue);

	return 0;
}


/* memory */

inline void* xp_malloc(int size)
{
	void	*ret;

	ret = malloc(size, M_DAZUKOBUF, M_WAITOK);

	if (ret == NULL)
		xp_print("dazuko: warning: malloc(%d, ...) -> NULL\n", size);

	return ret;
}

inline int xp_free(void *ptr)
{
	if (ptr == NULL)
	{
		xp_print("dazuko: warning: xp_free(NULL)\n");
		return 0;
	}

	free(ptr, M_DAZUKOBUF);

	return 0;
}


/* path attribute */

inline int xp_is_absolute_path(const char *path)
{
	if (path == NULL)
	{
		xp_print("dazuko: warning: xp_is_absolute_path(NULL)\n");
		return 0;
	}

	return (path[0] == '/');
}


/* atomic */

inline int xp_atomic_set(struct xp_atomic *atomic, int value)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_set(NULL, %d)\n", value);
		return -1;
	}

	atomic->c = (char)value;
	return 0;
}

inline int xp_atomic_inc(struct xp_atomic *atomic)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_inc(NULL)\n");
		return -1;
	}

	(atomic->c)++;
	return 0;
}

inline int xp_atomic_dec(struct xp_atomic *atomic)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_dec(NULL)\n");
		return -1;
	}

	(atomic->c)--;
	return 0;
}

inline int xp_atomic_read(struct xp_atomic *atomic)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_read(NULL)\n");
		return -1;
	}

	return atomic->c;
}


/* file descriptor */

inline int xp_copy_file(struct xp_file *dest, struct xp_file *src)
{
	if (dest == NULL || src == NULL)
	{
		xp_print("dazuko: error: xp_copy_file(%s, %s)\n", dest ? "dest" : "NULL", src ? "src" : "NULL");
		return -1;
	}

	dest->file = src->file;

	return 0;
}

inline int xp_compare_file(struct xp_file *file1, struct xp_file *file2)
{
	if (file1 == NULL || file2 == NULL)
	{
		xp_print("dazuko: error: xp_compare_file(%s, %s)\n", file1 ? "file1" : "NULL", file2 ? "file2" : "file2");
		return -1;
	}

	return (file1->file != file2->file);
}


/* file structure */

inline int xp_file_struct_check(struct dazuko_file_struct *dfs)
{
	int	i;

	if (dfs == NULL)
	{
		xp_print("dazuko: warning: xp_file_struct_check(NULL)\n");
		return -1;
	}

	for (i=0 ; fubyte((dfs->extra_data->user_filename)+i) ; i++);

	i++;

	dfs->filename = (char *)xp_malloc(i);
	if (dfs->filename == NULL)
		return -1;

	if (copyin(dfs->extra_data->user_filename, dfs->filename, i) != 0)
	{
		xp_print("dazuko: error: xp_file_struct_check.copyin() failed\n");
		xp_free(dfs->filename);
		return -1;
	}

	dfs->filename_length = i - 1;  /* the string length */

	return 0;
}

inline int xp_file_struct_check_cleanup(struct dazuko_file_struct *dfs)
{
	return 0;
}

static int dazuko_file_struct_cleanup(struct dazuko_file_struct **dfs)
{
	if (!dfs)
		return 0;

	if (!*dfs)
		return 0;

	xp_file_struct_check_cleanup(*dfs);

	if ((*dfs)->filename)
		xp_free((*dfs)->filename);

	if ((*dfs)->extra_data)
		xp_free((*dfs)->extra_data);

	xp_free(*dfs);

	*dfs = NULL;

	return 0;
}


/* system calls */

#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_open(struct proc *p, struct open_args *uap)
{
	struct dazuko_file_struct	*dfs = NULL;
	int				error = 0;
	int				check;
#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	struct xp_file			file;
#endif

	if (p == NULL)
	{
		xp_print("dazuko: warning: freebsd_dazuko_sys_open(NULL, ...)\n");
		check = 0;
	}
	else
	{
		check = dazuko_sys_open_check(freebsd_getpid(p));
	}

	if (check)
	{
		if (uap == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_open(p, NULL)\n");
			check = 0;
		}
		else if (p->p_cred == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_open.p->p_cred=NULL\n");
			check = 0;
		}
		else
		{
			dfs = (struct dazuko_file_struct *)xp_malloc(sizeof(struct dazuko_file_struct));
			if (dfs)
			{
				dazuko_bzero(dfs, sizeof(struct dazuko_file_struct));

				dfs->extra_data = (struct xp_file_struct *)xp_malloc(sizeof(struct xp_file_struct));
				if (dfs->extra_data)
				{
					dazuko_bzero(dfs->extra_data, sizeof(struct xp_file_struct));

					dfs->extra_data->user_filename = uap->path;

					error = dazuko_sys_open_pre(dfs, uap->flags, uap->mode, freebsd_getpid(p), p->p_cred->p_ruid);
				}
				else
				{
					xp_free(dfs);
					dfs = NULL;
				}
			}
		}
	}

	if (error)
	{
		/* access should be blocked */

		error = EPERM;
	}
	else
	{
		/* call the standard open function */
		error = original_sys_open(p, uap);

		#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
		{
			if (check)
			{
				if (error)
				{
					check = 0;
				}
				else if (p->p_fd == NULL)
				{
					check = 0;
				}
				else if (p->p_retval[0] < 0)
				{
					check = 0;
				}
				else if (p->p_retval[0] >= p->p_fd->fd_nfiles)
				{
					check = 0;
				}
				else if (p->p_fd->fd_ofiles == NULL)
				{
					check = 0;
				}
				else if (p->p_fd->fd_ofiles[p->p_retval[0]] == NULL)
				{
					check = 0;
				}
				else
				{
					file.file = p->p_fd->fd_ofiles[p->p_retval[0]];
				}
			}
		}
		#endif
	}

	if (check)
	{
		#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
		{
			if (!error && dfs)
				dazuko_sys_open_post(dfs, &file, uap->flags, uap->mode);
		}
		#endif
	}

	dazuko_file_struct_cleanup(&dfs);

	return error;
}
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_close(struct proc *p, struct close_args *uap)
{
	/* The kernel wants to close the given file
	 * descriptor. */

	int		error;
	int		check;
	struct xp_file	file;

	if (p == NULL)
	{
		xp_print("dazuko: warning: freebsd_dazuko_sys_close(NULL, ...)\n");
		check = 0;
	}
	else
	{
		check = dazuko_sys_close_check(freebsd_getpid(p));
	}

	if (check)
	{
		if (uap == NULL)
		{
			check = 0;
		}
		else if (p->p_fd == NULL)
		{
			check = 0;
		}
		else if (uap->fd < 0)
		{
			check = 0;
		}
		else if ((u_int)(uap->fd) >= p->p_fd->fd_nfiles)
		{
			check = 0;
		}
		else if (p->p_fd->fd_ofiles == NULL)
		{
			check = 0;
		}
		else if (p->p_fd->fd_ofiles[uap->fd] == NULL)
		{
			check = 0;
		}
		else
		{
			/* grab the file* for possible later use */
			file.file = p->p_fd->fd_ofiles[uap->fd];
		}
	}

	error = original_sys_close(p, uap);

	if (!error && check)
	{
		if (p->p_cred == NULL)
			xp_print("dazuko: error: freebsd_dazuko_sys_close.p->p_cred=NULL\n");
		else
			dazuko_sys_close_post(&file, 0, 0, freebsd_getpid(p), p->p_cred->p_ruid);
	}

	return error;
}
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
static int freebsd_dazuko_sys_write(struct proc *p, struct write_args *uap)
{
	/* The kernel wants to write to the given file
	 * descriptor. */

	int		num;
	int		check;
	struct xp_file	file;

	check = dazuko_sys_write_check();

	if (check && uap->fd >= 0 && uap->fd < NR_OPEN)
	{
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		{
			read_lock(&current->files->file_lock);
		}
		#endif

		/* Grab a copy of the file* "just in case*. */
		file.file = current->files->fd[fd];

		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		{
			read_unlock(&current->files->file_lock);
		}
		#endif
	}

	num = original_sys_write(p, uap);
	
	/* did we actually write anything? */

	if (num > 0)
		dazuko_sys_write_post(&file);

	return num;
}
#endif

#ifdef ON_EXEC_SUPPORT
static int freebsd_dazuko_sys_execve(struct proc *p, struct execve_args *uap)
{
	/* The kernel wants to execute the given file.
	 * Because the given structure contains stack
	 * address information, we can't simply call
	 * the default standard execve. Instead we
	 * have to manually inline the standard execve
	 * call. */

	struct dazuko_file_struct	*dfs = NULL;
	int				error = 0;
	int				check;

	check = dazuko_sys_exec_check();

	if (check)
	{
		if (p == NULL || uap == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_execve(%s, %s)\n", p ? "p" : "NULL", uap ? "uap" : "NULL");
			check = 0;
		}
		else if (p->p_cred == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_execve.p->p_cred=NULL\n");
			check = 0;
		}
		else
		{
			dfs = (struct dazuko_file_struct *)xp_malloc(sizeof(struct dazuko_file_struct));
			if (dfs)
			{
				dazuko_bzero(dfs, sizeof(struct dazuko_file_struct));

				dfs->extra_data = (struct xp_file_struct *)xp_malloc(sizeof(struct xp_file_struct));
				if (dfs->extra_data)
				{
					dazuko_bzero(dfs->extra_data, sizeof(struct xp_file_struct));

					dfs->extra_data->user_filename = uap->fname;

					error = dazuko_sys_exec_pre(dfs, 0, 0, freebsd_getpid(p), p->p_cred->p_ruid);

					dazuko_file_struct_cleanup(&dfs);
				}
				else
				{
					xp_free(dfs);
					dfs = NULL;
				}
			}
		}
	}

	if (error)
	{
		return error;
	}

	return original_sys_execve(p, uap);
}
#endif


/* system hook */

inline int xp_sys_hook()
{
	sdev = make_dev(&cdepts, 0, UID_ROOT, GID_WHEEL, 0600, DEVICE_NAME);

#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	DPRINT(("dazuko: hooked sys_open\n"));
	original_sys_open = sysent[SYS_open].sy_call;
	sysent[SYS_open] = freebsd_dazuko_sys_open_sysent;
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
        DPRINT(("dazuko: hooked sys_close\n"));
	original_sys_close = sysent[SYS_close].sy_call;
	sysent[SYS_close] = freebsd_dazuko_sys_close_sysent;
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
        DPRINT(("dazuko: hooked sys_write\n"));
	original_sys_write = sysent[SYS_write].sy_call;
	sysent[SYS_write] = freebsd_dazuko_sys_write_sysent;
#endif

#ifdef ON_EXEC_SUPPORT
        DPRINT(("dazuko: hooked sys_execve\n"));
	original_sys_execve = sysent[SYS_execve].sy_call;
	sysent[SYS_execve] = freebsd_dazuko_sys_execve_sysent;
#endif

	return 0;
}

inline int xp_sys_unhook()
{
#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	if (sysent[SYS_open].sy_call != freebsd_dazuko_sys_open_sysent.sy_call)
		xp_print("dazuko: open system call not correct (system may be left in an unstable state!)\n");
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	if (sysent[SYS_close].sy_call != freebsd_dazuko_sys_close_sysent.sy_call)
		xp_print("dazuko: close system call not correct (system may be left in an unstable state!)\n");
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
	if (sysent[SYS_write].sy_call != freebsd_dazuko_sys_write_sysent.sy_call)
		xp_print("dazuko: write system call not correct (system may be left in an unstable state!)\n");
#endif

#ifdef ON_EXEC_SUPPORT
	if (sysent[SYS_execve].sy_call != freebsd_dazuko_sys_execve_sysent.sy_call)
		xp_print("dazuko: execve system call not correct (system may be left in an unstable state!)\n");
#endif


#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	sysent[SYS_open].sy_call = (sy_call_t*)original_sys_open;
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	sysent[SYS_close].sy_call = (sy_call_t*)original_sys_close;
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
	sysent[SYS_write].sy_call = (sy_call_t*)original_sys_write;
#endif

#ifdef ON_EXEC_SUPPORT
	sysent[SYS_execve].sy_call = (sy_call_t*)original_sys_execve;
#endif

	destroy_dev(sdev);

	return 0;
}


/* output */

inline int xp_print(const char *fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	return 0;
}


/* ioctl's */

int freebsd_dazuko_device_open(dev_t dev, int oflags, int devtype, struct proc *p)
{
	DPRINT(("dazuko: freebsd_dazuko_device_open() [%d]\n", freebsd_getpid(p)));

	return 0;
}

int freebsd_dazuko_device_read(dev_t dev, struct uio *uio, int ioflag)
{
	/* Reading from the dazuko device simply
	 * returns the device number. This is to
	 * help out the daemon. */

	size_t	dev_major_len;
	char	tmp[20];

	if (uio == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read(..., NULL, ...)\n");
		return EFAULT;
	}

	if (uio->uio_procp == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read.uio->uio_procp=NULL\n");
		return EFAULT;
	}

	if (uio->uio_procp->p_cred == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read.uio->uio_procp->p_cred=NULL\n");
		return EFAULT;
	}

	DPRINT(("dazuko: freebsd_dazuko_device_read() [%d]\n", freebsd_getpid(uio->uio_procp)));

	/* non-root daemons are ignored */
	if (uio->uio_procp->p_cred->p_ruid != 0)
		return 0;

	if (dev_major < 0)
		return ENODEV;

	if (uio->uio_offset != 0)
		return 0;

	/* print dev_major to a string
 	* and get length */
	dazuko_bzero(tmp, sizeof(tmp));

	dev_major_len = snprintf(tmp, sizeof(tmp), "%d", dev_major);

	if (tmp[sizeof(tmp)-1] != 0)
	{
		xp_print("dazuko: failing device_read, device number overflow for dameon %d (dev_major=%d)\n", freebsd_getpid(uio->uio_procp), dev_major);
		return EFAULT;
	}

	if (uio->uio_iov == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read.uio->uio_iov=NULL\n");
		return EFAULT;
	}

	if (uio->uio_iov->iov_len < dev_major_len)
		return EINVAL;

	/* copy dev_major string to userspace */
	if (uiomove(tmp, dev_major_len, uio) != 0)
		return EFAULT;

	return 0;
}

int freebsd_dazuko_device_ioctl(dev_t dev, u_long cmd, caddr_t data, int fflag, struct proc *p)
{
	/* A daemon uses this function to interact with
	 * the kernel. A daemon can set scanning parameters,
	 * give scanning response, and get filenames to scan. */

	int			error = 0;
	int			len;
	int			opt;
	struct access_t		*u_dazuko;
	struct option_t		*u_option;
	struct slot_t		*s;
	char			*k_param;
	int			i;

	if (p == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_ioctl(..., NULL)\n");
		return EFAULT;
	}

	if (p->p_cred == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_ioctl.p->p_cred=NULL\n");
		return EFAULT;
	}

	/* non-root daemons are ignored */
	if (p->p_cred->p_ruid != 0)
		return 0;

	if (data == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_ioctl(..., NULL, ...)\n");
		return EFAULT;
	}

	/* A macro is used to translate the cmd argument.
	 * This keeps it compatible across various Linux
	 * platforms. */
	switch (cmd)
	{
		case IOCTL_GET_AN_ACCESS:
			/* The daemon is requesting a filename of a file
			 * to scan. This code will wait until a filename
			 * is available, or until we should be killed.
			 * (killing is done if any errors occur as well
			 * as when the user kills us) */

/*
			error = verify_area(VERIFY_WRITE, (void *)param, sizeof(struct access_t));
			if (error)
				return error;
*/

			u_dazuko = *(struct access_t **)data;

/* DOWN? */
			s = dazuko_get_an_access(freebsd_getpid(p));

			if (s == NULL)
				return EINTR;

/* DOWN */

			/* Slot IS in DAZUKO_WORKING state. Copy all the
			 * necessary information to userspace structure. */

			if (copyout(s->filename, u_dazuko->filename, (s->filenamelength)+1) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->event), &(u_dazuko->event), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->o_flags), &(u_dazuko->o_flags), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->o_mode), &(u_dazuko->o_mode), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->kuid), &(u_dazuko->uid), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->kpid), &(u_dazuko->pid), sizeof(int)) != 0)
			{
				error = EFAULT;
			}

			xp_up(&(s->mutex));
/* UP */

			if (error)
			{
				dazuko_state_error(s, DAZUKO_WORKING);
			}

			return error;

		case IOCTL_RETURN_ACCESS:
			/* The daemon has finished scanning a file
			 * and has the response to give. The daemon's
			 * slot should be in the DAZUKO_WORKING state. */

/*
			error = verify_area(VERIFY_READ, (void *)param, sizeof(struct access_t));
			if (error)
				return error;
*/

			u_dazuko = *(struct access_t **)data;

			/* copy the response into the slot */
			if (copyin(&(u_dazuko->deny), &i, sizeof(int)) != 0)
			{
				return EFAULT;
			}

			return dazuko_return_access(freebsd_getpid(p), i);

		case IOCTL_SET_OPTION:
			/* The daemon wants to set a configuration
			 * option in the kernel. */

/*
			error = verify_area(VERIFY_READ, (void *)param, 2*sizeof(int));
			if (error)
				return error;
*/

			u_option = *(struct option_t **)data;

			/* copy option type from userspace */
			if (copyin(&(u_option->command), &opt, sizeof(int)) != 0)
				return EPERM;

			/* copy path length from userspace */
			if (copyin(&(u_option->buffer_length), &len, sizeof(int)) != 0)
				return EPERM;

			/* sanity check */
			if (len < 0 || len > 1024)
				return EPERM;

/*
			error = verify_area(VERIFY_READ, u_opt->buffer, len);
			if (error)
				return error;
*/

			k_param = (char *)xp_malloc(len + 1);
			if (!k_param)
				return EFAULT;

			/* We must copy the param from userspace to kernelspace. */

			if (copyin(u_option->buffer, k_param, len) != 0)
			{
				xp_free(k_param);
				return EFAULT;
			}

			k_param[len] = 0;

			error = dazuko_set_option(freebsd_getpid(p), opt, k_param, len);

			xp_free(k_param);

			return error;

		default:
			xp_print("dazuko: daemon %d requested unknown device_ioctl %d (possible bug)\n", freebsd_getpid(p), _IO(dev_major, cmd));

			break;
	}

	return 0;
}

int freebsd_dazuko_device_close(dev_t dev, int fflag, int devtype, struct proc *p)
{
	return dazuko_unregister_daemon(freebsd_getpid(p));
}


/* init/exit */

static int dazuko_loader(struct module *m, int what, void *arg)
{
	int	err = 0;

	switch (what)
	{
		case MOD_LOAD:
			err = dazuko_init();
			break;

		case MOD_UNLOAD:
			if (dazuko_exit() != 0)
				err = EPERM;
			break;

		default:
			err = EINVAL;
			break;
	}

	return err;
}

DEV_MODULE(dazuko, dazuko_loader, NULL);
        
/*
//MODULE_AUTHOR("H+BEDV Datentechnik GmbH <freebsd_support@antivir.de>");
//MODULE_DESCRIPTION("allow 3rd-party file access control");
//#ifdef MODULE_LICENSE
//MODULE_LICENSE("GPL");
//#else
//static const char __module_license[] __attribute__((section(".modinfo"))) = "license=GPL";
//#endif   
                
//EXPORT_NO_SYMBOLS;

//#else
         
//module_init(freebsd_dazuko_init);
//module_exit(freebsd_dazuko_exit);
// module_init(int freebsd_dazuko_init(void));
// module_exit(void freebsd_dazuko_exit(void));

//#endif
*/

