/* DazukoXP. Allow cross platform file access control for 3rd-party applications.
   Written by John Ogness <jogness@antivir.de>

   Copyright (C) 2002-2003, H+BEDV Datentechnik GmbH
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. Neither the name of Dazuko nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

#include "dazuko_platform.h"

#include "dazuko_xp.h"
#include "dazukoio.h"

#define NUM_SLOT_LISTS	5
#define NUM_SLOTS	25

#define	SCAN_ON_OPEN		(access_mask & ON_OPEN)
#define	SCAN_ON_CLOSE		(access_mask & ON_CLOSE)
#define	SCAN_ON_EXEC		(access_mask & ON_EXEC)
#define	SCAN_ON_CLOSE_MODIFIED	(access_mask & ON_CLOSE_MODIFIED)

struct path_t
{
	/* A node in a linked list of paths. Used
	 * for the include and exclude lists. */

	struct path_t	*next;
	int		len;
	char		path[1];	/* this MUST be at the end of the struct */
};

struct hash_t
{
	/* A node in a linked list of filenames.
	 * Used for the list of files to be
	 * scanned on close. */

	struct hash_t	*next;
	struct xp_file	file;
	int		dirty;
	int		namelen;
	char		name[1];	/* this MUST be at the end of the struct */
};

struct slot_list_container_t
{
	struct slot_list_t	*slot_list;
	struct xp_mutex		mutex;
};

struct slot_list_t
{
	struct xp_atomic	use_count;
	struct slot_t		slots[NUM_SLOTS];
	char			reg_name[1];	/* this MUST be at the end of the struct */
};

struct one_slot_state_not_condition_param_t
{
	struct slot_t	*slot;
	int		state;
};

struct two_slot_state_not_condition_param_t
{
	struct slot_t	*slot1;
	int		state1;
	struct slot_t	*slot2;
	int		state2;
};

struct get_ready_slot_condition_param_t
{
	struct slot_t		*slot;
	struct slot_list_t	*slotlist;
};

static char				access_mask = 7;
static struct slot_list_container_t	slot_lists[NUM_SLOT_LISTS];
static struct path_t			*incl_paths = NULL;
static struct path_t			*excl_paths = NULL;
static struct hash_t			*hash = NULL;
static struct xp_rwlock			lock_hash;
static struct xp_rwlock			lock_lists;
static struct xp_atomic			active;

static struct xp_queue			wait_kernel_waiting_for_free_slot;
static struct xp_queue			wait_daemon_waiting_for_work;
static struct xp_queue			wait_kernel_waiting_while_daemon_works;
static struct xp_queue			wait_daemon_waiting_for_free;

inline void dazuko_bzero(void *p, int len)
{
	/* "zero out" len bytes starting with p */

	char	*ptr = (char *)p;

	while (len--)
		*ptr++ = 0;
}

static inline int dazuko_slot_state(struct slot_t *s)
{
	int state;

/* DOWN */
	if (xp_down(&(s->mutex)) != 0)
		return -EINTR;

	state = s->state;

	xp_up(&(s->mutex));
/* UP */

	return state;
}

static int one_slot_state_not_condition(void *param)
{
	return (dazuko_slot_state(((struct one_slot_state_not_condition_param_t *)param)->slot)
		!= ((struct one_slot_state_not_condition_param_t *)param)->state);
}

static int two_slot_state_not_condition(void *param)
{
	return (dazuko_slot_state(((struct two_slot_state_not_condition_param_t *)param)->slot1)
		!= ((struct two_slot_state_not_condition_param_t *)param)->state1
		&& dazuko_slot_state(((struct two_slot_state_not_condition_param_t *)param)->slot2)
		!= ((struct two_slot_state_not_condition_param_t *)param)->state2);
}

static inline int __dazuko_change_slot_state(struct slot_t *s, int from_state, int to_state)
{
	/* Make a predicted state transition. We fail if it
	 * is an unpredicted change. We can ALWAYS go to the
	 * to_state if it is the same as from_state. Not SMP safe! */

	if (to_state != from_state)
	{
		/* make sure this is a predicted transition and there
		 * is a daemon on this slot (pid != 0)*/
		if (s->state != from_state || s->pid == 0)
			return 0;
	}

	s->state = to_state;

	/* handle appropriate wake_up's for basic
	 * state changes */

	if (to_state == DAZUKO_READY)
	{
		xp_notify(&wait_kernel_waiting_for_free_slot);
	}
	else if (to_state == DAZUKO_FREE)
	{
		xp_notify(&wait_kernel_waiting_while_daemon_works);
		xp_notify(&wait_daemon_waiting_for_free);
	}

	return 1;
}

static int dazuko_change_slot_state(struct slot_t *s, int from_state, int to_state, int release)
{
	/* SMP safe version of __dazuko_change_slot_state().
	 * This should only be used if we haven't
	 * already aquired slot.mutex. Use this function
	 * with CAUTION, since the mutex may or may not
	 * be released depending on the return value AND
	 * on the value of the "release" argument. */

	int	success;

	/* if we are interrupted, report the state as unpredicted */
/* DOWN */
	if (xp_down(&(s->mutex)) != 0)
		return 0;

	success = __dazuko_change_slot_state(s, from_state, to_state);

	/* the mutex is released if the state change was
	 * unpredicted or if the called wants it released */
	if (!success || release)
		xp_up(&(s->mutex));
/* UP */
	return success;
}

static struct slot_t * _dazuko_find_slot(int pid, int release, struct slot_list_t *sl)
{
	/* Find the first slot with the same given
	 * pid number. SMP safe. Use this function
	 * with CAUTION, since the mutex may or may not
	 * be released depending on the return value AND
	 * on the value of the "release" argument. */

	int		i;
	struct slot_t	*s = NULL;

	if (sl == NULL)
	{
		xp_print("dazuko: invalid slot_list given (bug!)\n");
		return NULL;
	}

	for (i=0 ; i<NUM_SLOTS ; i++)
	{
		s = &(sl->slots[i]);
/* DOWN */
		/* if we are interrupted, we say that no
 		* slot was found */
		if (xp_down(&(s->mutex)) != 0)
			return NULL;

		if (s->pid == pid)
		{
			/* we release the mutex only if the
	 		* called wanted us to */
			if (release)
				xp_up(&(s->mutex));
/* UP */
			return s;
		}

		xp_up(&(s->mutex));
/* UP */
	}

	return NULL;
}

static struct slot_t * dazuko_find_slot_and_slotlist(int pid, int release, struct slot_list_t *slist, struct slot_list_t **sl_result)
{
	struct slot_t		*s;
	int			i;
	struct slot_list_t	*sl;

	if (slist == NULL)
	{
		for (i=0 ; i<NUM_SLOT_LISTS ; i++)
		{
/* DOWN */
			/* if we are interrupted, we say that no
 			* slot was found */
			if (xp_down(&(slot_lists[i].mutex)) != 0)
				return NULL;

			sl = slot_lists[i].slot_list;

			xp_up(&(slot_lists[i].mutex));
/* UP */

			if (sl != NULL)
			{
				s = _dazuko_find_slot(pid, release, sl);
				if (s != NULL)
				{
					/* set the current slot_list */
					if (sl_result != NULL)
						*sl_result = sl;

					return s;
				}
			}
		}
	}
	else
	{
		return _dazuko_find_slot(pid, release, slist);
	}

	return NULL;
}

static inline struct slot_t * dazuko_find_slot(int pid, int release, struct slot_list_t *slist)
{
	struct slot_list_t	*sl;

	return dazuko_find_slot_and_slotlist(pid, release, slist, &sl);
}

static int dazuko_insert_path_fs(struct path_t **list, char *fs_path, int fs_len)
{
	/* Create a new path_t structure and insert it
	 * into the linked list given (list argument).
	 * The fs_len argument is to help speed things
	 * up so we don't have to calculate the length
	 * of fs_path. The fs_path argument is in
	 * userspace! */

	struct path_t	*newitem;
	struct path_t	*tmp;

	/* create a new path_t structure making room for path also */
	newitem = (struct path_t *)xp_malloc(sizeof(struct path_t) + fs_len);
	if (!newitem)
		return -EFAULT;

	/* fs_path is already in kernelspace */
	memcpy(newitem->path, fs_path, fs_len);

	newitem->path[fs_len] = 0;

	while (newitem->path[fs_len] == 0)
	{
		fs_len--;
		if (fs_len == 0)
			break;
	}

	newitem->len = fs_len;

	/* we want only absolute paths */
	if (!xp_is_absolute_path(newitem->path))
	{
		xp_free(newitem);
		return -EINVAL;
	}

	/* check if this path already exists in the list */
	for (tmp=*list ; tmp ; tmp=tmp->next)
	{
		if (newitem->len == tmp->len)
		{
			if (memcmp(newitem->path, tmp->path, tmp->len) == 0)
			{
				/* we already have this path */

				xp_free(newitem);

				return 0;
			}
		}
	}

	DPRINT(("dazuko: adding %s %s\n", (list == &incl_paths) ? "incl" : "excl", newitem->path));

	/* add path_t to head of linked list */
/* LOCK */
	xp_write_lock(&lock_lists);
	newitem->next = *list;
	*list = newitem;
	xp_write_unlock(&lock_lists);
/* UNLOCK */

	return 0;
}

static void dazuko_remove_all_hash(void)
{
	/* Empty the hash linked list. */

	struct hash_t	*tmp;

/* LOCK */
	xp_write_lock(&lock_hash);
	while (hash)
	{
		tmp = hash;
		hash = hash->next;

		xp_free(tmp);
	}
	xp_write_unlock(&lock_hash);
/* UNLOCK */
}

static void dazuko_remove_all_paths(void)
{
	/* Empty both include and exclude path_t
	 * linked lists. */

	struct path_t	*tmp;

/* LOCK */
	xp_write_lock(&lock_lists);

	/* empty include paths list */
	while (incl_paths)
	{
		tmp = incl_paths;
		incl_paths = incl_paths->next;

		DPRINT(("dazuko: removing incl %s\n", tmp->path));

		xp_free(tmp);
	}

	/* empty exclude paths list */
	while (excl_paths)
	{
		tmp = excl_paths;
		excl_paths = excl_paths->next;

		DPRINT(("dazuko: removing excl %s\n", tmp->path));

		xp_free(tmp);
	}

	xp_write_unlock(&lock_lists);
/* UNLOCK */
}

int dazuko_unregister_daemon(int pid)
{
	/* We unregister the daemon by finding the
	 * slot with the same slot->pid as the the
	 * current process id, the daemon. */

	struct slot_t		*s;
	struct slot_list_t	*sl;

	DPRINT(("dazuko: dazuko_device_release() [%d]\n", pid));

	/* find our slot and hold the mutex
	 * if we find it */
/* DOWN? */
	s = dazuko_find_slot_and_slotlist(pid, 0, NULL, &sl);

	if (s == NULL)
	{
		xp_print("dazuko: daemon %d had no slot (possible bug)\n", pid);
		return -1;
	}

/* DOWN */

	/* clearing the pid makes the slot available */
	s->pid = 0;

	/* reset slot state */
	__dazuko_change_slot_state(s, DAZUKO_FREE, DAZUKO_FREE);

	xp_atomic_dec(&(sl->use_count));

	xp_up(&(s->mutex));
/* UP */

	/* active should always be positive here, but
	 * let's check just to be sure. ;) */
	if (xp_atomic_read(&active) > 0)
	{
		/* active and the kernel usage counter
		 * should always reflect how many daemons
		 * are active */

		xp_atomic_dec(&active);
	}
	else
	{
		xp_print("dazuko: active count error (possible bug)\n");
	}

	/* Wake up any kernel processes that are
	 * waiting for an available slot. Remove
	 * all the include and exclude paths
	 * if there are no more daemons */

	if (xp_atomic_read(&active) == 0)
	{
		/* clear out include and exclude paths */
		/* are we sure we want to do this? */
		dazuko_remove_all_paths();

		/* clear out hash nodes */
		dazuko_remove_all_hash();
	}

	xp_notify(&wait_kernel_waiting_for_free_slot);
	xp_notify(&wait_kernel_waiting_while_daemon_works);

	return 0;
}

inline int dazuko_state_error(struct slot_t *s, int current_state)
{
	if (dazuko_change_slot_state(s, current_state, DAZUKO_BROKEN, 1))
	{
		xp_notify(&wait_kernel_waiting_for_free_slot);
		xp_notify(&wait_kernel_waiting_while_daemon_works);
	}

	return 0;
}

static int dazuko_register_daemon(int pid, const char *reg_name, int string_length)
{
	const char		*p1;
	char			*p2;
	struct slot_t		*s;
	struct slot_list_t	*sl;
	int			i;

	DPRINT(("dazuko: dazuko_register_daemon() [%d]\n", pid));

	if (reg_name == NULL)
		return -EPERM;

	/* Find the slot_list with the matching name. */

	for (i=0 ; i<NUM_SLOT_LISTS ; i++)
	{
/* DOWN */
		/* if we are interrupted, we say that it
 		* was interrupted */
		if (xp_down(&(slot_lists[i].mutex)) != 0)
			return -EINTR;

		sl = slot_lists[i].slot_list;

		xp_up(&(slot_lists[i].mutex));
/* UP */

		if (sl != NULL)
		{
			p1 = reg_name;
			p2 = sl->reg_name;

			while (*p1 == *p2)
			{
				if (*p1 == 0)
					break;

				p1++;
				p2++;
			}

			if (*p1 == *p2)
				break;
		}
	}

	if (i == NUM_SLOT_LISTS)
	{
		/* There is no slot_list with this name. We
		 * need to make one. */

		sl = (struct slot_list_t *)xp_malloc(sizeof(struct slot_list_t) + string_length);
		if (!sl)	
			return -EFAULT;

		dazuko_bzero(sl, sizeof(struct slot_list_t) + string_length);
		xp_atomic_set(&(sl->use_count), 0);

		p1 = reg_name;
		p2 = sl->reg_name;

		while (*p1)
		{
			*p2 = *p1;

			p1++;
			p2++;
		}
		*p2 = 0;

		/* give each slot a unique id */
		for (i=0 ; i<NUM_SLOTS ; i++)
		{
			sl->slots[i].id = i;
			xp_init_mutex(&(sl->slots[i].mutex));
		}

		/* we need to find an empty slot */
		for (i=0 ; i<NUM_SLOT_LISTS ; i++)
		{
/* DOWN */
			/* if we are interrupted, we need to cleanup
 			* and return error */
			if (xp_down(&(slot_lists[i].mutex)) != 0)
			{
				xp_free(sl);
				return -EINTR;
			}

			if (slot_lists[i].slot_list == NULL)
			{
				slot_lists[i].slot_list = sl;

				xp_up(&(slot_lists[i].mutex));
/* UP */
				break;
			}

			xp_up(&(slot_lists[i].mutex));
/* UP */
		}

		if (i == NUM_SLOT_LISTS)
		{
			/* no empty slot :( */
			xp_free(sl);
			return -EBUSY;
		}
	}

	/* find a slot with pid 0 and hold the mutex
	* if we find one */
/* DOWN? */
	s = dazuko_find_slot(0, 0, sl);

	if (s == NULL)
		return -EBUSY;

/* DOWN */

	/* We have found a slot, so increment the active
	 * variable and the kernel module use counter.
	 * The module counter will always reflect the
	 * number of daemons. */

	xp_atomic_inc(&active);

	s->pid = pid;

	xp_atomic_inc(&(sl->use_count));

	/* the daemon is registered, but not yet
	 * ready to receive files */
	__dazuko_change_slot_state(s, DAZUKO_FREE, DAZUKO_FREE);

	DPRINT(("dazuko: slot[%d] assigned to daemon %d\n", s->id, pid));

	xp_up(&(s->mutex));
/* UP */

	return 0;
}

struct slot_t* dazuko_get_an_access(int pid)
{
	/* The daemon is requesting a filename of a file
	 * to scan. This code will wait until a filename
	 * is available, or until we should be killed.
	 * (killing is done if any errors occur as well
	 * as when the user kills us) */

	/* If a slot is returned, it will be already locked! */

	int						i;
	struct slot_t					*s;
	struct one_slot_state_not_condition_param_t	cond_p;

tryagain:
	/* find our slot */
	s = dazuko_find_slot(pid, 1, NULL);

	if (s == NULL)
	{
		i = dazuko_register_daemon(pid, "_COMPAT", 7);
		if (i != 0)
		{
			xp_print("dazuko: unregistered daemon %d attempted to get access\n", pid);
			return NULL;
		}

		s = dazuko_find_slot(pid, 1, NULL);
		if (s == NULL)
		{
			xp_print("dazuko: unregistered daemon %d attempted to get access\n", pid);
			return NULL;
		}

		xp_print("dazuko: warning: daemon %d is using a deprecated protocol\n", pid);
	}

	/* the daemon is now ready to receive a file */
	dazuko_change_slot_state(s, DAZUKO_READY, DAZUKO_READY, 1);

	cond_p.slot = s;
	cond_p.state = DAZUKO_READY;
	if (xp_wait_until_condition(&wait_daemon_waiting_for_work, one_slot_state_not_condition, &cond_p) != 0)
	{
		/* The user has issued an interrupt.
		 * Return an error. The daemon should
		 * unregister itself. */

		DPRINT(("dazuko: daemon %d killed while waiting for work\n", pid));

		if (dazuko_change_slot_state(s, DAZUKO_READY, DAZUKO_BROKEN, 1) || dazuko_change_slot_state(s, DAZUKO_WAITING, DAZUKO_BROKEN, 1))
		{
			xp_notify(&wait_kernel_waiting_for_free_slot);
			xp_notify(&wait_kernel_waiting_while_daemon_works);
		}

		return NULL;
	}

	/* slot SHOULD now be in DAZUKO_WAITING state */

	/* we will be working with the slot, so
	 * we need to lock it */
/* DOWN */
	if (xp_down(&(s->mutex)) != 0)
	{
		return NULL;
	}

	if (!__dazuko_change_slot_state(s, DAZUKO_WAITING, DAZUKO_WORKING))
	{
		/* State transition error. Try again., */

		xp_up(&(s->mutex));
/* UP */
		goto tryagain;
	}

	/* Slot IS in DAZUKO_WORKING state. Copy all the
	 * necessary information to userspace structure. */

	/* IMPORTANT: slot is still locked! */

	return s;  /* access is available */
}

int dazuko_return_access(int pid, int response)
{
	/* The daemon has finished scanning a file
	 * and has the response to give. The daemon's
	 * slot should be in the DAZUKO_WORKING state. */

	struct slot_t					*s;
	struct one_slot_state_not_condition_param_t	cond_p;

	/* find our slot */
	s = dazuko_find_slot(pid, 1, NULL);

	if (s == NULL)
	{
		/* It appears the kernel isn't interested
		 * in us or our response. It gave our slot away! */

		DPRINT(("dazuko: daemon %d unexpectedly lost slot\n", pid));

		return -EPERM;
	}

	/* we will be writing into the slot, so we
	 * need to lock it */
/* DOWN */
	if (xp_down(&(s->mutex)) != 0)
	{
		return -EINTR;
	}

	if (!__dazuko_change_slot_state(s, DAZUKO_WORKING, DAZUKO_DONE))
	{
		/* The slot is in the wrong state. We will
		 * assume the kernel has cancelled the file
		 * access. */

		DPRINT(("dazuko: response from daemon %d on slot[%d] not needed\n", pid, s->id));

		xp_up(&(s->mutex));
/* UP */
		return 0;
	}

	s->response = response;

	xp_up(&(s->mutex));
/* UP */

	/* wake up any kernel processes that are
	 * waiting for responses */
	xp_notify(&wait_kernel_waiting_while_daemon_works);

	cond_p.slot = s;
	cond_p.state = DAZUKO_DONE;
	if (xp_wait_until_condition(&wait_daemon_waiting_for_free, one_slot_state_not_condition, &cond_p) != 0)
	{
		/* The user has issued an interrupt.
		 * Return an error. The daemon should
		 * unregister itself. */

		DPRINT(("dazuko: daemon %d killed while waiting for response acknowledgement\n", pid));

		return -EINTR;
	}

	return 0;
}

int dazuko_set_option(int pid, int opt, void *param, int len)
{
	/* The daemon wants to set a configuration
	 * option in the kernel. */

	struct slot_t	*s;
	int		error;

	/* sanity check */
	if (len < 0 || len > 1024)
		return -EPERM;

	/* make sure we are already registered
	 * (or that we don't register twice) */

	/* find our slot */
	s = dazuko_find_slot(pid, 1, NULL);

	if (opt == REGISTER)
	{
		if (s != NULL)
		{
			/* We are already registered! */

			xp_print("dazuko: daemon %d already assigned to slot[%d]\n", pid, s->id);

			return -EPERM;
		}
	}
	else
	{
		if (s == NULL)
		{
			error = dazuko_register_daemon(pid, "_COMPAT", 7);
			if (error)
			{
				xp_print("dazuko: unregistered daemon %d attempted to get access\n", pid);
				return -EPERM;
			}

			s = dazuko_find_slot(pid, 1, NULL);
			if (s == NULL)
			{
				xp_print("dazuko: unregistered daemon %d attempted to get access\n", pid);
				return -EPERM;
			}

			xp_print("dazuko: warning: daemon %d is using a deprecated protocol\n", pid);
		}
	}

	/* check option type and take the appropriate action */
	switch (opt)
	{
		case SET_ACCESS_MASK:
			memcpy(&access_mask, (char *)param, sizeof(char));
			break;

		case ADD_INCLUDE_PATH:
			error = dazuko_insert_path_fs(&incl_paths, (char *)param, len);
			if (error)
				return error;
			break;

		case ADD_EXCLUDE_PATH:
			error = dazuko_insert_path_fs(&excl_paths, (char *)param, len);
			if (error)
				return error;
			break;

		case REGISTER:
			/* We register the daemon by finding an
			 * unused slot (slot->pid=0) and setting
			 * the slot->pid to the proccess id of
			 * the current proccess, the daemon. */

			error = dazuko_register_daemon(pid, (char *)param, len);
			if (error)
				return error;

			break;

		case REMOVE_ALL_PATHS:
			dazuko_remove_all_paths();
			break;

		default:
			xp_print("dazuko: daemon %d requested unknown set %d (possible bug)\n", pid, opt);
			break;
	}

	return 0;
}

static struct slot_t * dazuko_get_and_hold_ready_slot(struct slot_list_t *sl)
{
	/* This is a simple search to find a
	 * slot whose state is DAZUKO_READY. This means
	 * it is able to accept work. If a slot
	 * is found, the slot.mutex is held so
	 * it can be filled with work by the caller.
	 * It is the responsibility of the caller
	 * to RELEASE THE MUTEX. */

	int		i;
	struct slot_t	*s;

	for (i=0 ; i<NUM_SLOTS ; i++)
	{
		s = &(sl->slots[i]);
/* DOWN? */
		if (dazuko_change_slot_state(s, DAZUKO_READY, DAZUKO_WAITING, 0))
		{
/* DOWN */
			return s;
		}
	}

	/* we didn't find a slot that is ready for work */

	return NULL;
}

int get_ready_slot_condition(void *param)
{
	return ((((struct get_ready_slot_condition_param_t *)param)->slot = dazuko_get_and_hold_ready_slot(((struct get_ready_slot_condition_param_t *)param)->slotlist)) != NULL
		|| xp_atomic_read(&active) == 0
		|| xp_atomic_read(&(((struct get_ready_slot_condition_param_t *)param)->slotlist->use_count)) == 0);
}

static int dazuko_run_daemon_on_slotlist(int event, char *filename, int filenamelength, int o_flags, int o_mode, struct slot_list_t *sl, int kpid, int kuid)
{
	/* This is the main function called by the kernel
	 * to work with a daemon. */

	int						rc;
	int						pid;
	struct slot_t					*s;
	struct get_ready_slot_condition_param_t		cond_p1;
	struct two_slot_state_not_condition_param_t	cond_p2;

begin:
	/* we initialize the slot value because
	 * we cannot guarentee that it will be
	 * assigned a new value BEFORE !active
	 * is checked */
	s = NULL;

	/* wait for a slot to become ready */
	cond_p1.slotlist = sl;
	cond_p1.slot = s;
	if (xp_wait_until_condition(&wait_kernel_waiting_for_free_slot, get_ready_slot_condition, &cond_p1) != 0)
	{
		/* The kernel process was killed while
		 * waiting for a slot to become ready.
		 * This is fine. */

		DPRINT(("dazuko: kernel process %d killed while waiting for free slot\n", kpid));

		return -1;  /* user interrupted */
	}

	/* Make sure we have a slot. We may have
	 * gotten past the last wait because we
	 * are no longer active. */

	s = cond_p1.slot;

	if (s == NULL)
	{
		/* We were no longer active. We don't
		 * need to initiate a daemon. This also
		 * means we never acquired the lock. */

		return 0;  /* allow access */
	}

/* DOWN */

	/* the slot is already locked at this point */

	/* grab the daemon's pid */
	pid = s->pid;

	/* At this point we have a locked slot. It IS
	 * sitting in the DAZUKO_WAITING state, waiting for
	 * us to give it some work. */
	
	/* set up the slot to do work */
	s->filename = filename;
	s->event = event;
	s->response = 0;
	s->kuid = kuid;
	s->kpid = kpid;
	s->o_flags = o_flags;
	s->o_mode = o_mode;
	s->filenamelength = filenamelength;

	/* we are done modifying the slot */
	xp_up(&(s->mutex));
/* UP */

	/* wake up any daemons waiting for work */
	xp_notify(&wait_daemon_waiting_for_work);

	/* wait until the daemon is finished with the slot */
	cond_p2.slot1 = s;
	cond_p2.state1 = DAZUKO_WAITING;
	cond_p2.slot2 = s;
	cond_p2.state2 = DAZUKO_WORKING;
	if (xp_wait_until_condition(&wait_kernel_waiting_while_daemon_works, two_slot_state_not_condition, &cond_p2) != 0)
	{
		/* The kernel process was killed while
		 * waiting for a daemon to process the file.
		 * This is fine. */

		DPRINT(("dazuko: kernel process %d killed while waiting for daemon response\n", kpid));

		/* change the slot's state to let the
		 * daemon know we are not interested
		 * in a response */
		dazuko_change_slot_state(s, DAZUKO_FREE, DAZUKO_FREE, 1);

		return -1;  /* user interrupted */
	}

	/* we are working with the slot, so
	 * we need to lock it */
/* DOWN */
	if (xp_down(&(s->mutex)) != 0)
	{
		return -1;  /* user interrupted */
	}

	/* make sure this is the right daemon */
	if (s->pid != pid)
	{
		/* This is a different daemon than
		 * the one we assigned work to.
		 * We need to scan again. */
		xp_up(&(s->mutex));
/* UP */
		goto begin;
	}

	/* The slot should now be in the DAZUKO_DONE state. */
	if (!__dazuko_change_slot_state(s, DAZUKO_DONE, DAZUKO_FREE))
	{
		/* The daemon was killed while scanning.
		 * We need to scan again. */

		xp_up(&(s->mutex));
/* UP */
		goto begin;
	}

	/* grab the response */
	rc = s->response;

	xp_up(&(s->mutex));
/* UP */

	/* CONGRATULATIONS! You successfully completed a full state cycle! */

	return rc;
}

static int dazuko_run_daemon(int event, char *filename, int filenamelength, int o_flags, int o_mode, int pid, int uid)
{
	struct slot_list_t	*sl;
	int			i;
	int			rc = 0;
	int			error;

	for (i=0 ; i<NUM_SLOT_LISTS ; i++)
	{
/* DOWN */
		/* if we are interrupted, we report error */
		if (xp_down(&(slot_lists[i].mutex)) != 0)
			return -EINTR;

		sl = slot_lists[i].slot_list;

		xp_up(&(slot_lists[i].mutex));
/* UP */

		if (sl != NULL)
		{
			error = dazuko_run_daemon_on_slotlist(event, filename, filenamelength, o_flags, o_mode, sl, pid, uid);

			if (error < 0)
			{
				/* most likely user interrupt */
				rc = error;
				break;
			}
			else if (error > 0)
			{
				/* this daemon wants access blocked */
				rc = 1;
			}
		}
	}

	return rc;
}

static inline int dazuko_is_our_daemon(int pid)
{
	/* Check if the current process is one
	 * of the daemons. */

	return (dazuko_find_slot(pid, 1, NULL) != NULL);
}

static int dazuko_is_selected(char *filename, int len)
{
	/* Check if the given filename (with path) is
	 * under our include directories but not under
	 * the exclude directories. */

	struct path_t	*path;

	/* If we are interrupted here, we will report that
	 * this file is not selected. This will make the
	 * kernel allow normal access. Is this dangerous? */
/* LOCK */
	xp_read_lock(&lock_lists);

	/* check if filename is under our include paths */
	for (path=incl_paths ; path ; path=path->next)
	{
		/* the include item must be at least as long as the given filename */
		if (path->len < len)
		{
			/* the include item should match the beginning of the given filename */
			if (memcmp(path->path, filename, path->len) == 0)
				break;
		}
	}

	/* If we didn't find a path, it isn't in our
	 * include directories. It can't be one of
	 * the selected files to scan. */
	if (!path)
	{
		xp_read_unlock(&lock_lists);
/* UNLOCK */
		return 0;
	}

	/* check if filename is under our exclude paths */
	for (path=excl_paths ; path ; path=path->next)
	{
		/* the exclude item must be at least as long as the given filename */
		if (path->len < len)
		{
			/* the exclude item should match the beginning of the given filename */
			if (memcmp(path->path,filename,path->len) == 0)
				break;
		}
	}

	xp_read_unlock(&lock_lists);
/* UNLOCK */

	/* If we got a path, then we are supposed
	 * to exclude this file for scanning. */
	if (path)
		return 0;

	/* if we made it this far, it is a selected file to scan */

	return 1;
}

static int dazuko_add_hash(struct xp_file *file, char *filename, int len)
{
	/* Add the given file and filename to the linked list
	 * of files to scan once they are closed. */

	struct hash_t	*h;

	/* create a new hash_t structure making room for name also */
	h = (struct hash_t *)xp_malloc(sizeof(struct hash_t) + len);
	if (!h)	
		return -EFAULT;

	/* fill in structure items */

	xp_copy_file(&(h->file), file);
	h->dirty = 0;
	h->namelen = len;
	memcpy(h->name, filename, len);
	h->name[len] = 0;

	/* add the new hash_t item to the head of the
	 * hast_t linked list */

/* LOCK */
	xp_write_lock(&lock_hash);
	h->next = hash;
	hash = h;
	xp_write_unlock(&lock_hash);
/* UNLOCK */
	return 0;
}

/* Code based on code from: Swade 12/08/02: Move dirty to end of list */
static void dazuko_mark_hash_dirty(struct xp_file *file)
{
	struct hash_t	*h = NULL;
	struct hash_t	*entry = NULL;
	struct hash_t	*prev = NULL;
	struct hash_t	*prev_entry = NULL;

/* LOCK */
	xp_write_lock(&lock_hash);

	for (h=hash ; h ; h=h->next)
	{
		/* not found if hit first dirty entry */
		if (h->dirty)
		{
			entry = NULL;
			break;
		}

		if (xp_compare_file(&(h->file), file) == 0)
		{
			/* we found the entry */

			prev_entry = prev;
			entry = h;
			break;
		}

		prev = h;
	} 

	if (entry)
	{
		if (!entry->dirty)
		{
			/* mark as dirty */
			entry->dirty = 1;

			/* If we already are last entry or next
		 	 * entry dirty, we don't need to move */

			if (entry->next)
			{
				if (!entry->next->dirty)
				{
					for (h=entry->next ; h ; h=h->next)
					{
						if (h->dirty)
							break;

						prev = h;
					}

					/* remove from current position */
					if (prev_entry)
						prev_entry->next = entry->next;
					else
						hash = entry->next;

					if (prev == NULL)
					{
						/* insert as first item */
						entry->next = hash;
						hash = entry;
					}
					else if (h)
					{
						/* insert before h (after prev) */
						entry->next = prev->next;
						prev->next = entry;
					}
					else
					{
						/* insert as last item (after prev) */
						entry->next = NULL;
						prev->next = entry;
					}
				}
			}
		}
	}

	xp_write_unlock(&lock_hash);
/* UNLOCK */

}

static struct hash_t *dazuko_get_hash(struct xp_file *file)
{
	/* Find the given file within our list
	 * and then remove it from the list and
	 * return it. */

	struct hash_t	*prev;
	struct hash_t	*cur;

/* LOCK */
	xp_write_lock(&lock_hash);

	prev = NULL;
	cur = hash;
	while (cur)
	{
		if (xp_compare_file(&(cur->file), file) == 0)
		{
			/* we found the entry */

			/* remove the item from the list */
			if (!prev)
				hash = cur->next;
			else
				prev->next = cur->next;
			break;
		}

		prev = cur;
		cur = cur->next;
	}

	xp_write_unlock(&lock_hash);
/* UNLOCK */

	return cur;
}

inline int dazuko_get_filename_length(char *filename)
{
	/* Get the length of the filename. There is
	 * currently a DAZUKO_FILENAME_MAX_LENGTH maximum size restriction
	 * on filenames. :( */

	int len;

	for (len=0 ; len<DAZUKO_FILENAME_MAX_LENGTH && filename[len]; len++);

	if (len == DAZUKO_FILENAME_MAX_LENGTH)
	{
		xp_print("dazuko: filename too long (%s)\n", filename);

		filename[DAZUKO_FILENAME_MAX_LENGTH] = 0;
	}

	return len;
}

static int dazuko_should_scan(struct dazuko_file_struct *kfs)
{
	/* Check if we are supposed to scan this file.
	 * This checks for all the correct file types,
	 * permissions, and if it is within the desired
	 * paths to scan. */

	int result = 0;

	/* check if we already know if we scan this file */
	switch (kfs->should_scan)
	{
		/* case 0 means that we do not know yet. This is a little
		 * confusing, because 0 represents uninitialized. However,
		 * the should_scan variable is used in this function ONLY
		 * so this optimization shouldn't cause any problems. */

		case 1:
			/* we already know it should be scanned */
			return 1;

		case 2:
			/* we already know it should not be scanned */
			return 0;
	}

	/* make necessary platform-dependent checks */
	if (xp_file_struct_check(kfs) == 0)
	{
		if (dazuko_is_selected(kfs->filename, kfs->filename_length))
		{
			/* If we made it this far, we are supposed
			 * to scan this file. We mark it so that
			 * any further immediate inquiries don't have
			 * to do all this work all over again. */
 
			/* yes, should be scanned */
			kfs->should_scan = 1;

			result = 1;
		}
		else
		{
			/* We will still mark it so that any further
			 * immediate inquiries don't have to do all
			 * this work all over again. */

			/* no, should not be scanned */
			kfs->should_scan = 2;
		}
	}

	xp_file_struct_check_cleanup(kfs);

	return result;
}

inline int dazuko_sys_exec_check()
{
	if (!SCAN_ON_EXEC)
		return 0;

	if (xp_atomic_read(&active) <= 0)
		return 0;

	return 1;
}

inline int dazuko_sys_exec_pre(struct dazuko_file_struct *kfs, int o_flags, int o_mode, int pid, int uid)
{
	int	error = 0;

	/* make sure we should scan this file */
	if (dazuko_should_scan(kfs))
		error = dazuko_run_daemon(ON_EXEC, kfs->filename, kfs->filename_length, o_flags, o_mode, pid, uid);

	if (error > 0)
	{
		/* access has been blocked */

		return -EPERM;
	}
	else if (error < 0)
	{
		/* user interrupted */

		return -EINTR;
	}

	return error;
}

inline int dazuko_sys_open_check(int pid)
{
	if (!SCAN_ON_OPEN && !SCAN_ON_CLOSE && !SCAN_ON_CLOSE_MODIFIED)
		return 0;

	if (xp_atomic_read(&active) <= 0)
		return 0;

	if (dazuko_is_our_daemon(pid))
		return 0;

	return 1;
}

inline int dazuko_sys_open_pre(struct dazuko_file_struct *kfs, int o_flags, int o_mode, int pid, int uid)
{
	/* The kernel wants to open the given filename
	 * with the given flags and mode. The dazuko_file_struct
	 * is used to handle the tricky job of cleaning
	 * up the many pieces of memory that may or may
	 * not be allocated. */

	int	error = 0;

	if (SCAN_ON_OPEN)
	{
		/* make sure we should scan this file */
		if (dazuko_should_scan(kfs))
		{
			error = dazuko_run_daemon(ON_OPEN, kfs->filename, kfs->filename_length, o_flags, o_mode, pid, uid);
		}
	}

	if (error > 0)
	{
		/* access will be blocked */

		/* dazuko_sys_open_post should NOT be called! */

		return -EPERM;
	}
	else if (error < 0)
	{
		/* user interrupted */

		/* dazuko_sys_open_post should NOT be called! */

		return -EINTR;
	}

	/* access allowed */

	return error;
}

inline int dazuko_sys_open_post(struct dazuko_file_struct *kfs, struct xp_file *file, int o_flags, int o_mode)
{
	/* If dazuko_sys_open_pre is called and returned 0,
	 * then this function MUST also be called. If fopen
	 * gave an error, call this function with a NULL
	 * file argument. */

	/* if the file was opened and we are interested
	 * in scanning on close, add this file to our hash_t list */

	if ((xp_atomic_read(&active) > 0) && file != NULL)
	{
		if (SCAN_ON_CLOSE || SCAN_ON_CLOSE_MODIFIED)
		{
			/* make sure we should scan this file */
			if (dazuko_should_scan(kfs))
				dazuko_add_hash(file, kfs->filename, kfs->filename_length);
		}
	}

	return 0;
}

inline int dazuko_sys_close_check(int pid)
{
	if (!SCAN_ON_CLOSE && !SCAN_ON_CLOSE_MODIFIED)
		return 0;

	if (xp_atomic_read(&active) <= 0)
		return 0;

	if (dazuko_is_our_daemon(pid))
		return 0;

	return 1;
}

inline int dazuko_sys_close_post(struct xp_file *file, int o_flags, int o_mode, int pid, int uid)
{
	/* The kernel has closed the given file
	 * descriptor. */

	struct hash_t	*h = NULL;

	if (file == NULL)
		return -1;

	/* find hash entry and remove it from list */
	h = dazuko_get_hash(file);

	/* if we found the file in our list and the file was
	 * successfully closed, we need to scan it */
	if (h)
	{
		/* determine if we are scanning on close and/or close_modified */

		/* note that modified has priority over just close */

		if (SCAN_ON_CLOSE_MODIFIED && h->dirty)
			dazuko_run_daemon(ON_CLOSE_MODIFIED, h->name, h->namelen, o_flags, o_mode, pid, uid);
		else if (SCAN_ON_CLOSE)
			dazuko_run_daemon(ON_CLOSE, h->name, h->namelen, o_flags, o_mode, pid, uid);

		/* clean up the hash_t structure */
		xp_free(h);
	}

	return 0;
}

inline int dazuko_sys_write_check()
{
	if (!SCAN_ON_CLOSE_MODIFIED)
		return 0;

	if (xp_atomic_read(&active) <= 0)
		return 0;

	return 1;
}

inline int dazuko_sys_write_post(struct xp_file *file)
{
	/* The kernel wants to write to the given file
	 * descriptor. */

	if (file == NULL)
		return -1;

	/* if we actually wrote something and we found the
	 * file in our list, set it as dirty */

	/* Swade 4/24/02: Move to end of clean list */
	dazuko_mark_hash_dirty(file);

	return 0;
}

inline int dazuko_init(void)
{
	int	i;
	int	error;

	xp_init_rwlock(&lock_hash);
	xp_init_rwlock(&lock_lists);

	xp_init_queue(&wait_kernel_waiting_for_free_slot);
	xp_init_queue(&wait_daemon_waiting_for_work);
	xp_init_queue(&wait_kernel_waiting_while_daemon_works);
	xp_init_queue(&wait_daemon_waiting_for_free);

	dazuko_bzero(&slot_lists, sizeof(slot_lists));

	for (i=0 ; i<NUM_SLOT_LISTS ; i++)
		xp_init_mutex(&(slot_lists[i].mutex));

	xp_atomic_set(&active, 0);

	error = xp_sys_hook();

	if (error == 0)
		xp_print("dazuko: loaded, version=%s\n", VERSION);

  	return error;
}

inline int dazuko_exit(void)
{
	int	error;
	int	i;

	i = xp_atomic_read(&active);

	if (i != 0)
	{
		xp_print("dazuko: warning: trying to remove Dazuko with %d process%s still registered\n", i, i==1 ? "" : "es");
		return -1;
	}

	dazuko_remove_all_paths();
	dazuko_remove_all_hash();

	error = xp_sys_unhook();

	if (error == 0)
	{
		for (i=0 ; i<NUM_SLOT_LISTS ; i++)
		{
			if (slot_lists[i].slot_list != NULL)
			{
				if (xp_atomic_read(&(slot_lists[i].slot_list->use_count)) != 0)
					xp_print("dazuko: slot_list count was not 0 (possible bug)\n");

				xp_free(slot_lists[i].slot_list);
				slot_lists[i].slot_list = NULL;
			}
		}

		xp_print("dazuko: unloaded, version=%s\n", VERSION);
	}

	return error;
}
